/*
//supercollider test code
OSCFunc.trace(true, true);
n= NetAddr("192.168.4.1", 9998).sendMsg(\ping)
n.sendMsg(\start)
n.sendMsg(\stop)
//or serial
*/

// ESP32-WROOM-32D
// pins 0, 4, 16 configures no WiFi, use an existing WiFi or set up a softAP

#include <Arduino.h>
#include <WiFi.h>
#include <WiFiUdp.h>
#include <OSCMessage.h>
#include <OSCData.h>

//--settings
#define INPORT 9998                  // EDIT osc input port
#define OUTPORT 57120                // EDIT osc output port (supercollider)
#define UPDATERATE 20                // milliseconds between each reading
const char *ssid = "myssid";         // EDIT wifi network name
const char *ssid2 = "myssid2";       // EDIT softAccessPoint network name
const char *password = "pass";       // EDIT password
const char *hostname = "handskar2";  // should be unique to each esp32
#define NUMPINS_ANA 5
#define NUMPINS_DIG 5
const uint8_t pins_ana[] = {36, 39, 34, 35, 32}; // 36 and 39 are labelled VP, VN
const uint8_t pins_dig[] = {25, 26, 27, 14, 12};
const uint8_t pins_config[] = {0, 4, 16};
const uint8_t pin_led = 2;

//--variables
uint8_t data_ana[NUMPINS_ANA];
uint8_t data_dig[NUMPINS_DIG];
bool running = false;
WiFiUDP Udp;
IPAddress outIP;
IPAddress myIP;
enum configuration {
  NOWIFI,
  WIFI,
  SOFTAP
};
configuration config;

void setup() {
  delay(100);
  for (uint8_t i = 0; i < NUMPINS_DIG; i++) {
    pinMode(pins_dig[i], INPUT_PULLDOWN);
  }
  pinMode(pins_config[0], INPUT_PULLUP);
  pinMode(pins_config[1], OUTPUT);
  pinMode(pins_config[2], INPUT_PULLUP);
  digitalWrite(pins_config[1], 0);
  if (digitalRead(pins_config[0]) == 0) {
    config = WIFI;
  } else if (digitalRead(pins_config[2]) == 0) {
    config = SOFTAP;
  } else {
    config = NOWIFI;
  }
  pinMode(pin_led, OUTPUT);

  Serial.begin(57600);

  if (config == WIFI) {
    WiFi.setHostname(hostname);
    WiFi.mode(WIFI_STA);
    WiFi.config(INADDR_NONE, INADDR_NONE, INADDR_NONE, INADDR_NONE);
    WiFi.begin(ssid, password);
    while (WiFi.status() != WL_CONNECTED) {
      digitalWrite(pin_led, 1);
      delay(250);
      digitalWrite(pin_led, 0);
      delay(250);
    }
    outIP = WiFi.localIP();
    outIP[3] = 255; // broadcast
    Udp.begin(INPORT);
  } else if (config == SOFTAP) {
    WiFi.softAP(ssid2, password);
    myIP = WiFi.softAPIP();
    Udp.begin(INPORT);
  }
}

void ping(OSCMessage &msg) { // reply with a pong
  outIP = Udp.remoteIP();
  OSCMessage outmsg("/pong");
  outmsg.add(myIP[0]);
  outmsg.add(myIP[1]);
  outmsg.add(myIP[2]);
  outmsg.add(myIP[3]);
  outmsg.add(hostname);
  Udp.beginPacket(outIP, OUTPORT);
  outmsg.send(Udp);
  Udp.endPacket();
}

void start(OSCMessage &msg) {
  outIP = Udp.remoteIP();
  running = true;
}

void stop(OSCMessage &msg) {
  outIP = Udp.remoteIP();
  running = false;
}

void loop() {
  static unsigned long nextTime = 0;
  if (millis() >= nextTime) {
    digitalWrite(pin_led, 0);

    //--osc in
    if (config != NOWIFI) {
      int packetSize = Udp.parsePacket();
      if (packetSize) {
        OSCMessage oscMsg;
        while (packetSize--) {
          oscMsg.fill(Udp.read());
        }
        if (!oscMsg.hasError()) {
          oscMsg.dispatch("/ping", ping);
          oscMsg.dispatch("/start", start);
          oscMsg.dispatch("/stop", stop);
        }
      }
    }

    //--analog inputs
    boolean changed_ana = false;
    for (uint8_t i = 0; i < NUMPINS_ANA; i++) {
      uint8_t val = analogRead(pins_ana[i]) >> 4; // resolution 12 down to 8 bits
      if (val != data_ana[i]) {
        data_ana[i] = val;
        changed_ana = true;
      }
    }
    if (changed_ana) {
      digitalWrite(pin_led, 1);

      //--serial out
      Serial.write(200);
      Serial.write(201);
      for (uint8_t i = 0; i < NUMPINS_ANA; i++) {
        Serial.write(data_ana[i]);
      }
      Serial.write(202);

      //--osc out
      if (running && (config != NOWIFI)) {
        OSCMessage msg("/pins");
        for (uint8_t i = 0; i < NUMPINS_ANA; i++) {
          msg.add(data_ana[i]);
        }
        Udp.beginPacket(outIP, OUTPORT);
        msg.send(Udp);
        Udp.endPacket();
        yield();
        msg.empty();
      }
    }

    //--digital inputs
    boolean changed_dig = false;
    for (uint8_t i = 0; i < NUMPINS_DIG; i++) {
      uint8_t val = digitalRead(pins_dig[i]);
      if (val != data_dig[i]) {
        data_dig[i] = val;
        changed_dig = true;
      }
    }
    if (changed_dig) {
      digitalWrite(pin_led, 1);

      //--serial out
      Serial.write(100);
      Serial.write(101);
      for (uint8_t i = 0; i < NUMPINS_DIG; i++) {
        Serial.write(data_dig[i]);
      }
      Serial.write(102);

      //--osc out
      if (running && (config != NOWIFI)) {
        OSCMessage msg("/pins_dig");
        for (uint8_t i = 0; i < NUMPINS_DIG; i++) {
          msg.add(data_dig[i]);
        }
        Udp.beginPacket(outIP, OUTPORT);
        msg.send(Udp);
        Udp.endPacket();
        yield();
        msg.empty();
      }
    }

    nextTime = millis() + UPDATERATE; // make sure not to read too often
  }
}
